#include <aiousb.h>
#include <stdio.h>
#include <unistd.h>

#include <mysql/mysql.h>

using namespace AIOUSB;


/* Helpful conversion constants. */
static const unsigned usec_per_sec = 1000000;
static const unsigned usec_per_msec = 1000;

/* These functions are written to match the win32
   signatures and behavior as closely as possible.
*/
bool QueryPerformanceFrequency(int64_t *frequency)
{
    /* Sanity check. */
    assert(frequency != NULL);

    /* gettimeofday reports to microsecond accuracy. */
    *frequency = usec_per_sec;

    return true;
}

bool QueryPerformanceCounter(int64_t *performance_count)
{
    struct timeval time;

    /* Sanity check. */
    assert(performance_count != NULL);

    /* Grab the current time. */
    gettimeofday(&time, NULL);
    *performance_count = time.tv_usec + /* Microseconds. */
                         time.tv_sec * usec_per_sec; /* Seconds. */

    return true;
}


#define NUM_INSERTIONS 6000
//#define ROWS_PER_INSERT 100

#define SQL_INSERT_HEADER "INSERT INTO DAQ_Data.RAW_DATA( TIMESTAMP, SGp01, SGp02, SGp03, SGp04, SGp05, SGp06, SGb1a, SGb1b, SGb1c, SGb2a, SGb2b, SGb2c, SGb3a, SGb3b, SGb3c, SGb4a, SGb4b, SGb4c, SGb5a, SGb5b, SGb5c, SGb6a, SGb6b, SGb6c, SGb01, SGb02, SGb03, SGb04, SGb05, SGb06, SGb07, SGb08, SGs01, SGs02, SGs03, SGs04, SGs05, SGs06, SGs07, SGs08, SGs09, SGs10, SGs11, SGs12, SGs13, SGs14, SGs15, SGs16, SGs17, SGs18, SGs19, SGs20, SGs21, SGs22, SGs23, SGs24, Temp1, Temp2, Acc1x, Acc1y, Acc1z, Acc2x, Acc2y, Acc2z) VALUES "

#define SQL_INSERT_FORMAT "( %ld, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f)"

size_t build_insert_string (double values[128], size_t dest_length, char *dest, bool final)
{
    size_t ret_val = 0; 
    int i;
    

    //ret_val = snprintf(dest, dest_length, "%s", SQL_INSERT_HEADER);

    ret_val += snprintf(dest + ret_val, dest_length - ret_val, SQL_INSERT_FORMAT "%s", time(NULL),
                                                values[0],
                                                values[1],
                                                values[2],
                                                values[3],
                                                values[4],
                                                values[5],
                                                values[6],
                                                values[7],
                                                values[8],
                                                values[9],
                                                values[10],
                                                values[11],
                                                values[12],
                                                values[13],
                                                values[14],
                                                values[15],
                                                values[16],
                                                values[17],
                                                values[18],
                                                values[19],
                                                values[20],
                                                values[21],
                                                values[22],
                                                values[23],
                                                values[24],
                                                values[25],
                                                values[26],
                                                values[27],
                                                values[28],
                                                values[29],
                                                values[30],
                                                values[31],
                                                values[32],
                                                values[33],
                                                values[34],
                                                values[35],
                                                values[36],
                                                values[37],
                                                values[38],
                                                values[39],
                                                values[40],
                                                values[41],
                                                values[42],
                                                values[43],
                                                values[44],
                                                values[45],
                                                values[46],
                                                values[47],
                                                values[48],
                                                values[49],
                                                values[50],
                                                values[51],
                                                values[52],
                                                values[53],
                                                values[54],
                                                values[55],
                                                values[56],
                                                values[57],
                                                values[58],
                                                values[59],
                                                values[60],
                                                values[61],
                                                values[62],
                                                values[63],
                                                final ? ";" : ",");
    return ret_val;
}

void usage ()
{
    printf ("sample -c <rows_per_insert>\n");
    printf ("eg: \"sample -c 100\n\"");
    printf ("defautls to 100\n");
}

//Return usec since mark. assumes mark is in the past and really recent. Will
// probably break around four seconds on a 32bit system
int usec_elapsed (clockid_t clock, struct timespec *mark)
{
    struct timespec now;
    clock_gettime(clock, &now);

    if (mark->tv_sec != now.tv_sec)
    {
        now.tv_nsec = (now.tv_sec - mark->tv_sec) * 1000000000 + now.tv_nsec;
    }

    return (int)((now.tv_nsec - mark->tv_nsec)/1000);


    return 0;
}

int main( int argc, char **argv ) {
    int CAL_CHANNEL = 5;
    int MAX_CHANNELS = 128;
    int NUM_CHANNELS = 64;
    unsigned short counts[ MAX_CHANNELS ];
    double volts[ MAX_CHANNELS ];
    unsigned char gainCodes[ NUM_CHANNELS ];
    uint64_t serialNumber;
    unsigned long result;
    unsigned long deviceMask;
    int MAX_NAME_SIZE = 20;
    char name[ MAX_NAME_SIZE + 2 ];
    unsigned long productID, nameSize, numDIOBytes, numCounters;
    unsigned long deviceIndex = 0;
    bool deviceFound = false;
    ADConfigBlock configBlock = {0};
    char *calibration_type;
    int rows_per_commit = 100;
    int opt;
    struct timespec clock_res;

    while ((opt = getopt (argc, argv, "c:")) != -1)
    {
        switch(opt)
        {
            case 'c':
                rows_per_commit = strtol(optarg, NULL, 0);
                break;
            case '?':
            default:
                usage();
                exit(1);
                break;
        };
    }

    printf("rows_per_commit = %d\n", rows_per_commit);

    if (clock_getres(CLOCK_MONOTONIC, &clock_res))
    {
        printf("Ettor getting clock resolution\n");
    }

    printf("Clock resolution: tv_sec = %ld, tv_nsec = %ld\n", clock_res.tv_sec, clock_res.tv_nsec);

    {
        int i;
        struct timespec current_time;
        struct timespec start_time;
        clock_gettime(CLOCK_MONOTONIC, &start_time);
        for (i = 0; i < 20; i++)
        {
            clock_gettime(CLOCK_MONOTONIC, &current_time);
            printf("Current time: tv_sec = %ld, tv_nsec = %ld\n", current_time.tv_sec, current_time.tv_nsec);
            usleep(50000);
            printf("usec_elapsed, %d\n", usec_elapsed(CLOCK_MONOTONIC, &current_time));
        }
    }


    //exit(0);

    printf("USB-AI16-16A sample program version %s, %s\n"
           "  This program demonstrates controlling a USB-AI16-16A device on\n"
           "  the USB bus. For simplicity, it uses the first such device found\n"
           "  on the bus.\n", 
           AIOUSB_GetVersion(), 
           AIOUSB_GetVersionDate()
           );

    /*
     * MUST call AIOUSB_Init() before any meaningful AIOUSB functions;
     * AIOUSB_GetVersion() above is an exception
     */
    result = AIOUSB_Init();
    if ( result != AIOUSB_SUCCESS ) {
        printf( "AIOUSB_Init() error: %d\n" , (int)result);
        goto out_main;
    }

    deviceMask = GetDevices(); /** call GetDevices() to obtain "list" of devices found on the bus */
    if ( !deviceMask  ) {
        printf( "No ACCES devices found on USB bus\n" );
        goto out_after_init;
    }

    /*
     * at least one ACCES device detected, but we want one of a specific type
     */
    AIOUSB_ListDevices(); /** print list of all devices found on the bus */

    while( deviceMask != 0 ) {
        if( ( deviceMask & 1 ) != 0 ) {
            // found a device, but is it the correct type?
            nameSize = MAX_NAME_SIZE;
            result = QueryDeviceInfo( deviceIndex, &productID, &nameSize, name, &numDIOBytes, &numCounters );
            if( result == AIOUSB_SUCCESS ) {
                if( productID >= USB_AI16_16A && productID <= USB_AIO12_128E ) {
                    deviceFound = true;                    // found a USB-AI16-16A family device
                    break;
                }
            } else
                printf( "Error '%s' querying device at index %lu\n", 
                        AIOUSB_GetResultCodeAsString( result ), 
                        deviceIndex 
                        );
        }
        deviceIndex++;
        deviceMask >>= 1;
    }

    if (!deviceFound ) {
        printf( "Failed to find USB-AI16-16A device\n" );
        goto out_after_init;
    }

    AIOUSB_Reset( deviceIndex );
    result = GetDeviceSerialNumber( deviceIndex, &serialNumber );

    if( result == AIOUSB_SUCCESS )
        printf( "Serial number of device at index %llx\n", (long long)serialNumber  );
    else
        printf( "Error '%s' getting serial number of device at index %lu\n",
                AIOUSB_GetResultCodeAsString( result ),
                deviceIndex );

    /*
     * demonstrate A/D configuration; there are two ways to configure the A/D;
     * one way is to create an ADConfigBlock instance and configure it, and then
     * send the whole thing to the device using ADC_SetConfig(); the other way
     * is to use the discrete API functions such as ADC_SetScanLimits(), which
     * send the new settings to the device immediately; here we demonstrate the
     * ADConfigBlock technique; below we demonstrate use of the discrete functions
     */
    AIOUSB_InitConfigBlock( &configBlock, deviceIndex, AIOUSB_FALSE );

    AIOUSB_SetAllGainCodeAndDiffMode( &configBlock, AD_GAIN_CODE_10V, AIOUSB_FALSE );
    ADCConfigBlockSetCalMode( &configBlock, AD_CAL_MODE_NORMAL );
    ADCConfigBlockSetTriggerMode( &configBlock, 0 );
    AIOUSB_SetScanRange( &configBlock, 0, 63 );
    ADCConfigBlockSetOversample( &configBlock, 0 );

    // memset(configBlock.registers,0,20 );
    // configBlock.registers[0x11] = 0x04;
    // configBlock.registers[0x12] = 0xf0;
    // configBlock.registers[0x13] = 0x0e;

    result = ADC_SetConfig( deviceIndex, configBlock.registers, &configBlock.size );

    if ( result != AIOUSB_SUCCESS ) {
        printf( "Error '%s' setting A/D configuration\n", 
                AIOUSB_GetResultCodeAsString( result ) 
                );
        goto out_after_init;
    }
    printf( "A/D settings successfully configured\n" );

    // calibration_type = strdup(":1TO1:");
    calibration_type = strdup(":AUTO:");

    result = ADC_SetCal( deviceIndex, calibration_type ); /**demonstrate automatic A/D calibration */

    if( result == AIOUSB_SUCCESS ) {
        printf( "Automatic calibration completed successfully\n" );
    } else {
        printf( "Error '%s' performing automatic A/D calibration\n", AIOUSB_GetResultCodeAsString( result ) );
        goto out_after_init;
    }

    /*
     * verify that A/D ground calibration is correct
     */
    ADC_SetOversample( deviceIndex, 0 );
    ADC_SetScanLimits( deviceIndex, CAL_CHANNEL, CAL_CHANNEL );
    ADC_ADMode( deviceIndex, 0 , AD_CAL_MODE_GROUND );

    // result = ADC_GetScan( deviceIndex, counts );

    // if( result == AIOUSB_SUCCESS )
    //     printf( "Ground counts = %u (should be approx. 0)\n", counts[ CAL_CHANNEL ] );
    // else
    //     printf( "Error '%s' attempting to read ground counts\n", 
    //             AIOUSB_GetResultCodeAsString( result ) );




    /*
     * verify that A/D reference calibration is correct
     */
    // ADC_ADMode( deviceIndex, 0 , AD_CAL_MODE_REFERENCE );
    // result = ADC_GetScan( deviceIndex, counts );
    // if( result == AIOUSB_SUCCESS )
    //     printf( "Reference counts = %u (should be approx. 65130)\n", counts[ CAL_CHANNEL ] );
    // else
    //     printf( "Error '%s' attempting to read reference counts\n", 
    //             AIOUSB_GetResultCodeAsString( result ) );
    /*
     * demonstrate scanning channels and measuring voltages
     */
    for( int channel = 0; channel < NUM_CHANNELS; channel++ )
        gainCodes[ channel ] = AD_GAIN_CODE_0_10V;

    ADC_RangeAll( deviceIndex, gainCodes, AIOUSB_TRUE );
    ADC_SetOversample( deviceIndex, 10 );
    ADC_SetScanLimits( deviceIndex, 0, NUM_CHANNELS - 1 );
    ADC_ADMode( deviceIndex, 0 /* TriggerMode */, AD_CAL_MODE_NORMAL );

    {//<Customer specific sample>
#define FIRST_SAMPLES 10
        int insert_length = 0x1000 * rows_per_commit;
        char *insert_statement = (char *)malloc(insert_length); //During testing 100 rows took about 70k bytes so this seems like more than enough
        MYSQL *con = mysql_init(NULL);
        int64_t time_start, time_end;
        int i;
        double samples[128];
        int committed = 0;
        struct timespec last_sample, remainder;
        int sleep_for;
        time_t timestamps[rows_per_commit];
        int current_length;

        printf("start sample\n");


        if (NULL == mysql_real_connect(con, NULL, "daq", "D@QUs3r01", NULL, 3306, NULL, 0))
        {
            printf("Unable to connect to db\n");
        }

        strcpy(insert_statement, SQL_INSERT_HEADER);
        current_length = strlen(insert_statement);

        QueryPerformanceCounter(&time_start);

        for (i = 0; i < NUM_INSERTIONS; i++)
        {
            //printf("In for loop");
             clock_gettime(CLOCK_MONOTONIC, &last_sample);
             result = ADC_GetScanV( deviceIndex, samples );
            // timestamps[i %rows_per_commit] = time(NULL);

            if ( (i + 1) % rows_per_commit == 0)
            {
                build_insert_string(samples, insert_length - current_length, &insert_statement[current_length], true);
                //printf("%s\n", insert_statement);
                committed += rows_per_commit;
                if (mysql_query(con, insert_statement))
                {
                    printf("Failed query\n");
                }
                strcpy(insert_statement, SQL_INSERT_HEADER);
                current_length = strlen(insert_statement);
            }
            else
            {
                current_length += build_insert_string(samples, insert_length - current_length, &insert_statement[current_length], false);
            }
            

            sleep_for = 10000 - usec_elapsed(CLOCK_MONOTONIC, &last_sample);
            if (sleep_for > 0)
            {
                // last_sample.tv_sec = 0;
                // last_sample.tv_nsec = 10000000;
                // nanosleep(&last_sample, &remainder);
                usleep(sleep_for > 0 ? sleep_for : 0);
            }
            else
            {
                //printf("Warning: loop took too long for desired speed\n");
             }
            
        }

        QueryPerformanceCounter(&time_end);
        mysql_close(con);

        printf("rows_per_commit = %d\n", rows_per_commit);
        printf("time_end - time_start = %lld\n", time_end - time_start);
        printf("committed = %d\n", committed);

        free(insert_statement);

        // if( result == AIOUSB_SUCCESS ) {
        //     printf( "Volts read:\n" );
        //     for( int channel = 0; channel < NUM_CHANNELS; channel++ )
        //         printf( "  Channel %2d = %f\n", channel, volts[ channel ] );
        // } else
        //     printf( "Error '%s' performing A/D channel scan\n", AIOUSB_GetResultCodeAsString( result ) );
    }//</Customer specific sample>
    /*
     * demonstrate reading a single channel in volts
     */
    // result = ADC_GetChannelV( deviceIndex, CAL_CHANNEL, &volts[ CAL_CHANNEL ] );
    // if( result == AIOUSB_SUCCESS )
    //     printf( "Volts read from A/D channel %d = %f\n", CAL_CHANNEL, volts[ CAL_CHANNEL ] );
    // else
    //     printf( "Error '%s' reading A/D channel %d\n", 
    //             AIOUSB_GetResultCodeAsString( result ), 
    //             CAL_CHANNEL );


/*
 * MUST call AIOUSB_Exit() before program exits,
 * but only if AIOUSB_Init() succeeded
 */
 out_after_init:
    AIOUSB_Exit();

 out_main:
    
    return ( int ) result;
}
